import logging
from colchian import Colchian
import unittest
import typing


class TestJsonTyping(unittest.TestCase):
    def setUp(self) -> None:
        logging.basicConfig(level=logging.WARNING)

    def tearDown(self) -> None:
        pass

    def test_version_defined(self):
        try:
            from colchian import __version__
        except ImportError:
            self.fail('__version__ not in package')

    def test_required(self):
        td = {
            'required': str,
            'not required': typing.Union[None, str],
            'also not required': typing.Optional[str],
            'not required but there': typing.Optional[str]
        }
        data = {
            'required': 'x',
            'not required but there': None
        }
        self.assertEqual(data, Colchian.validated(data, td),
                         'required and not required fields work with Union[None, type] and Optional[type]')
        data = {
        }
        with self.assertRaises(Colchian.ValidationError,
                               msg='missing required key in data results in Colchian.ValidationError'):
            Colchian.validated(data, td)

    def test_simple_types(self):
        td = {
            'i': int,
            'f': float,
            's': str,
            'b': bytes,
            't': bool
        }
        data = {
            'i': 1,
            'f': 1.5,
            's': 'test',
            'b': b'test',
            't': True
        }
        self.assertEqual(data, Colchian.validated(data, td),
                         'simple types check out')
    def test_not_simple_types(self):
        td = {
            'i': int,
            'f': float,
        }
        data = {
            'i': 1,
            'f': '1.5'
        }
        with self.assertRaises(Colchian.ValidationError, msg='simple types mismatches'):
            Colchian.validated(data, td, strict=True)  # default
        try:
            Colchian.validated(data, td, strict=False)
        except Colchian.ValidationError:
            self.fail("simple types cast when not strict")
        data = {
            'i': 1,
            'f': 'banana'
        }
        with self.assertRaises(Colchian.ValidationError, msg='simple types mismatches when not strict and no cast'):
            Colchian.validated(data, td, strict=False)

    def test_wildcards(self):
        td = {
            '*:1': {
                'i': int
            },
            '*:2': str,
            'a': float
        }
        data = {
            'x': {'i': 1},
            'y': 'some test',
            'a': 1.2,
            'z': {'i': 123}
        }
        self.assertEqual(data, Colchian.validated(data, td),
                         'wilcards match unmatched values')

    def test_nested_types(self):
        td = {
            'i': int,
            'd': {
                'i': int,
                's': str,
                'd': {
                    'f': float
                }
            }
        }
        data = {
            'i': 1,
            'd': {
                'i': 2,
                's': 'test',
                'd': {
                    'f': 3.0
                }
            }
        }
        self.assertEqual(data, Colchian.validated(data, td),
                         'nested types check out')

    def test_nested_wildcards(self):
        td = {
            'x': {
                '*': {
                    'i': int
                }
            },
            'y': {
                '*:1': {
                    's': str
                },
                '*:2': {
                    'f': float
                }
            }
        }
        data = {
            'x': {
                'a': {'i': 1},
                'b': {'i': 2}
            },
            'y': {
                'a': {'s': 'pi'},
                'b': {'f': .1},
                'c': {'f': 3.14},
                'd': {'s': 'text'}
            }
        }
        self.assertEqual(data, Colchian.validated(data, td),
                         'nested wildcards get correctly resolved')

    def test_custom_type(self):
        class MyInt(int):
            pass
        td = {
            'x': MyInt
        }
        data = {
            'x': 3
        }
        with self.assertRaises(Colchian.ValidationError, msg='custom types fail when strict'):
            Colchian.validated(data, td)
        try:
            result = Colchian.validated(data, td, strict=False)
            self.assertIsInstance(result['x'], MyInt, 'custom type preserved')
        except Colchian.ValidationError:
            self.fail('custom types only validate when not strict')

    def _test_union(self, td):
        data = {
            'x': 3
        }
        try:
            result = Colchian.validated(data, td)
            self.assertEqual(3, result['x'], 'int option resolved correctly for union')
        except Colchian.ValidationError:
            self.fail('int option allowed in union')
        data = {
            'x': 'test'
        }
        try:
            result = Colchian.validated(data, td)
            self.assertEqual('test', result['x'], 'str option resolved correctly for union')
        except Colchian.ValidationError:
            self.fail('str option allowed in union')
        data = {
            'x': 0.3
        }
        with self.assertRaises(Colchian.ValidationError, msg='float option not allowed in str, int union'):
            Colchian.validated(data, td)

    def test_union_variants(self):
        # helper tries int, str, and float value for 'x'
        self._test_union({
            'x': typing.Union[int, str]
        })

        self._test_union({
            'x': (int, str)
        })

        self._test_union({
            'x': (None, typing.Union[int, str])
        })

        td = {
            'x': typing.Union[int, str, None]
        }
        try:
            Colchian.validated({}, td)
        except Colchian.ValidationError:
            self.fail('optional value allowed for None in Union')

        td = {
            'x': typing.Optional[typing.Union[int, str]]
        }
        try:
            Colchian.validated({}, td)
        except Colchian.ValidationError:
            self.fail('optional value allowed for Optional Union')

        td = {
            'x': (None, typing.Union[int, str])
        }
        try:
            Colchian.validated({}, td)
        except Colchian.ValidationError:
            self.fail('optional value allowed for Union paired with None')

    def test_callable(self):
        def yes_no(x, strict, keys):
            if x not in ['yes', 'no']:
                raise Colchian.ValidationError('Only yes or no allowed')
            return True if x == 'yes' else False

        def add_params(x, y, z, strict, keys):
            return x + y + z

        self.assertEqual({'x': True}, Colchian.validated({'x': 'yes'}, {'x': yes_no}),
                         'basic callable returning bool')

        self.assertEqual({'y': True}, Colchian.validated({'y': 'yes'}, {'y': yes_no}),
                         'basic callable returning bool')

        with self.assertRaises(Colchian.ValidationError, msg='strict bool fails on non-bool text'):
            Colchian.validated({'bool': 'not True'}, {'bool': Colchian.text_bool})
        try:
            self.assertEqual({'bool': True}, Colchian.validated({'bool': 'True'}, {'bool': Colchian.text_bool},
                                                                strict=False),
                             'non-bool text is True when not strict')
        except Colchian.ValidationError:
            self.fail('non-bool text does not fail when not strict')
        with self.assertRaises(Colchian.ValidationError, msg='callable fails on mismatched text'):
            Colchian.validated({'x': 'test'}, {'x': yes_no})

        self.assertEqual({'x': 'xab'}, Colchian.validated({'x': 'x'}, {'x': (add_params, 'a', 'b')}),
                         'basic callable returning bool')

    def test_list(self):
        td = {
            'xs': (None, [int]),
            'ys': (None, [int]),
            'zs': (None, [str])
        }
        data = {
            'xs': [0, 1, 2],
            'ys': [],
            'zs': ['x']
        }
        self.assertEqual(data, Colchian.validated(data, td),
                         'basic lists get correctly resolved')
        with self.assertRaises(Colchian.ValidationError, msg='only int accepted in list of int when strict'):
            Colchian.validated({'xs': ['3']}, td)
        self.assertEqual({'xs': [1, 2, 3]}, Colchian.validated({'xs': [1.0, 2, '3']}, td, strict=False),
                         'list int values are cast when not strict')
        with self.assertRaises(Colchian.ValidationError, msg='only str accepted in list of str when strict'):
            Colchian.validated({'zs': [3]}, td)
        self.assertEqual({'zs': ['1.0', '2', '3']}, Colchian.validated({'zs': [1.0, 2, '3']}, td, strict=False),
                         'list str values are cast when not strict')
        self.assertEqual({'xs': [1, 2, 3]},
                         Colchian.validated({'xs': [1.0, 2, '3']}, {'xs': typing.List[int]}, strict=False),
                         'list int values are cast when not strict')

    def test_values(self):
        self.assertEqual({'x': 'a', 'y': 'c'}, Colchian.validated({'x': 'a', 'y': 'c'}, {'x': ('a', 'b'), 'y': 'c'}),
                         'values get correctly resolved')

    def test_type_key(self):
        self.assertEqual({1: 'one'}, Colchian.validated({1: 'one'}, {int: 'one'}),
                         'type keys are validated')
        with self.assertRaises(Colchian.ValidationError, msg='type key mismatch caught'):
            Colchian.validated({'1': 'one'}, {int: 'one'})
        with self.assertRaises(Colchian.ValidationError, msg='type key value mismatch caught'):
            Colchian.validated({1: 'one'}, {int: 'two'})

    def test_callable_key(self):
        def uppercase(x, strict, keys):
            if x.upper() != x:
                raise Colchian.ValidationError('key not in uppercase')
            return x

        def uppercase_no_keywords(x):
            if x.upper() != x:
                raise Colchian.ValidationError('key not in uppercase')
            return x

        self.assertEqual({'ABC': 1}, Colchian.validated({'ABC': 1}, {uppercase: int}),
                         'callable keys are validated')
        with self.assertRaises(Colchian.ValidationError, msg='callable key mismatch caught'):
            Colchian.validated({'Abc': 1}, {uppercase: int})
        with self.assertRaises(Colchian.ValidationError, msg='callable key value mismatch caught'):
            Colchian.validated({'ABC': 'one'}, {uppercase: int})
        with self.assertRaises(Colchian.ValidationError, msg='same result with keywords lacking'):
            Colchian.validated({'ABC': 'one'}, {uppercase_no_keywords: int})

    def test_callable_parameters_key(self):
        def greater(x, y, strict, keys):
            if x <= y:
                raise Colchian.ValidationError(f'{x} not greater than {y}')
            return x
        self.assertEqual({4: 1}, Colchian.validated({4: 1}, {(greater, 3): (greater, 0)}),
                         'callable keys with parameters are validated')
        with self.assertRaises(Colchian.ValidationError, msg='callable key with parameters mismatch caught'):
            Colchian.validated({4: 1}, {(greater, 5): int})

    def test_tuple_restricted_key(self):
        self.assertEqual({3: 1, 4: 1}, Colchian.validated({3: 1, 4: 1}, {(3, 4): 1}),
                         'tuple restricted keys are validated')
        with self.assertRaises(Colchian.ValidationError, msg='tuple restricted key with parameters mismatch caught'):
            Colchian.validated({3: 1, 5: 1}, {(3, 4): 1})
        self.assertEqual({1: 1, 10: 1}, Colchian.validated({1: 1, 10: 1}, {tuple(range(1, 11)): 1}),
                         'tuple restricted keys are validated')

    def test_constructor_override(self):
        class MyDict(dict):
            important = True

        def my_dict_factory(x: MyDict):
            result = type(x)()
            result.important = x.important
            return result

        class MyDict2(MyDict):
            important = True
            also_important = True

        Colchian.type_factories[MyDict] = my_dict_factory

        md = Colchian.validated(MyDict({'a': 1}), {'a': int})
        self.assertEqual(True, md.important, 'attribute on overridden class is set')

        md = MyDict({'a': 1})
        md.important = False
        md = Colchian.validated(md, {'a': int})
        self.assertEqual(False, md.important, 'attribute on overridden class is set correctly')

        md = MyDict2({'a': 1})
        md.important = False
        md = Colchian.validated(md, {'a': int})
        self.assertEqual(False, md.important, 'attribute on overridden class is set correctly on child class')
        self.assertIsInstance(md, MyDict2, 'correct constructor is used in construction validated result')

    def test_key_type(self):
        td = {
            int: int,
        }
        self.assertEqual({1: 2}, Colchian.validated({1: 2}, td), 'int key check works')
        with self.assertRaises(Colchian.ValidationError, msg='cast not allowed when strict'):
            Colchian.validated({'1': 2}, td, strict=True)
        self.assertEqual({1: 2}, Colchian.validated({1: 2}, td, strict=False), 'int cast works when not strict')
        with self.assertRaises(Colchian.ValidationError, msg='impossible cast not allowed') as cm:
            Colchian.validated({'1.0': 2}, td, strict=False)
        self.assertEqual(
            'could not match to only wildcard <class \'int\'>, raised `key `["1.0"]` cannot be cast to <class \'int\'>`',
            str(cm.exception)
        )

    def test_key_type_tuple(self):
        td = {
            (int, str): int,
        }
        self.assertEqual({1: 2}, Colchian.validated({1: 2}, td, strict=True),
                         'int key allowed for (int, str) pair')
        self.assertEqual({'test': 2}, Colchian.validated({'test': 2}, td, strict=True),
                         'str key allowed for (int, str) pair')
        td = {
            (str, int): int,
        }
        self.assertEqual({1: 2}, Colchian.validated({1: 2}, td, strict=True),
                         'no conversion to first type in pair')
        with self.assertRaises(Colchian.ValidationError, msg='float key not allowed for strict (int, str)') as cm:
            Colchian.validated({1.0: 2}, td, strict=True)
        self.assertEqual({1: 2}, Colchian.validated({1: 2}, td, strict=True),
                         'no conversion to first type in pair')
        self.assertEqual({'1.0': 2}, Colchian.validated({1.0: 2}, td, strict=False),
                         'conversion to first type (str, int) if not strict')
        td = {
            (int, str): int,
        }
        self.assertEqual({1: 2}, Colchian.validated({1.0: 2}, td, strict=False),
                         'conversion to first type (int, str) if not strict')



import os
import re
from setuptools import setup

__name__ = 'colchian'

project_urls = {
    'Home page': 'https://pypi.org/project/colchian',
    'Source Code': 'https://github.com/jaapvandervelde/colchian',
    'Documentation': 'https://colchian.readthedocs.io/'
}

version_fn = os.path.join(__name__, "_version.py")
__version__ = "unknown"
try:
    version_line = open(version_fn, "rt").read()
except EnvironmentError:
    pass  # no version file
else:
    version_regex = r"^__version__ = ['\"]([^'\"]*)['\"]"
    m = re.search(version_regex, version_line, re.M)
    if m:
        __version__ = m.group(1)
    else:
        print(f'unable to find version in {version_fn}')
        raise RuntimeError(f'If {version_fn} exists, it is required to be well-formed')

with open("README.md", "r") as rm:
    long_description = rm.read()

setup(
    name=__name__,
    packages=['colchian'],
    version=__version__,
    license='MIT',
    description='Validate json/yaml documents using a Python dictionary defining keys and types.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='BMT Commercial Australia Pty Ltd, Jaap van der Velde',
    author_email='jaap.vandervelde@bmtglobal.com',
    url='https://gitlab.com/jaapvandervelde/colchian',
    keywords=['json', 'validator'],
    install_requires=[],
    extras_require={
        'dev': [
            'mkdocs',
            'pymdown-extensions',
            'jinja2<3.1.0'
        ]
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.7',
    ]
)



from logging import getLogger
from typing import Union, List, Any
from copy import copy


class Colchian:
    class ValidationError(Exception):
        ...

    """
    Colchian is a library for validating and coercing data structures. You don't typically instantiate this class, but
    instead use the static and class methods to validate data.

    Examples:
    ```python
    from colchian import Colchian

    type_dict = {"an integer": int, "some strings": [str]}
    data = {"an integer": 42, "some strings": ["vastly", "hugely", "mind-bogglingly", "big"]}
    valid_data = Colchian.validated(data, type_dict)
    ```
    """
    ALLOW_INSPECT = True
    """
    If set to `True`, allows Colchian to inspect callables to avoid passing `strict` or `keys` arguments if the 
    callable does not support these. If you set `ALLOW_INSPECT` to `False`, this may improve performance, but then
    all callables must support `strict` and `keys` arguments getting passed to them.
    
    Example:
    ```
    from colchian import Colchian
    
    
    def the_answer(x):
        return x == 42
    
    
    print(Colchian.validated({'answer': 42}, {'answer': the_answer}))
    
    
    def the_answer_no_inspect(x, strict, keys):
        return x == 42
    
    
    Colchian.ALLOW_INSPECT = False
    print(Colchian.validated({'answer': 42}, {'answer': the_answer_no_inspect}))
    ```
    These function identically, but the second one is faster, because it doesn't inspect the callable - it does
    however require the unused `strict` and `keys` arguments. 
    """

    logger = getLogger()

    type_factories = {}
    """
    A dictionary of `type: factory_function` pairs. The factory function is called when a new instance of the type
    needs to be constructed, allowing you to set certain defaults. The factory function must accept a single argument,
    which will be the value of the data being validated. The factory function must return an instance of the type of
    that data, with initialisation or other operations performed on it.
    
    Example:
    ```
    from colchian import Colchian    
    
    class MyDict(dict):
        def __init__(self, p):
            super().__init__()
            self.p = p
            self.script = None
    
    
    def my_dict(d):
        result = type(d)(d.p)  # setting `p` on the result to `d.p`
        result.script = __file__
        return result
        
   
    d = MyDict('42')
    Colchian.type_factories[MyDict] = my_dict  # setting up the factory function for the type
    d2 = Colchian.validated(d, {})
    print(d2.script, d2.p == d2.p)  # print the name of the script that called validated(), and True 
    ``` 
    """

    @staticmethod
    def format_keys(keys: List[Any]) -> str:
        """
        Formats a list of keys into a string for use in error messages

        Args:
            keys: keys of any type (though typically str), that must convert into str

        Returns:
            a string representation of how `keys` would be used to access a dict, in backticks

        Examples:
        ```
        >>> Colchian.format_keys(['a', 'b', 'c'])
        '`["a"]["b"]["c"]`'
        ```
        """
        keys = "".join(
            f"[\"{key}\"]" if i == 0 else f"[\"{key}\"]"
            if isinstance(key, str) and len(key) > 0 and key[0] + key[-1] != "[]"
            else str(key) for i, key in enumerate(keys))
        return f'`{keys}`'

    @classmethod
    def text_bool(cls, x: Any, strict: bool, keys: List[str]) -> bool:
        """
        Usable example of a function that can be used to represent a boolean as case-insensitive text.
        Valid representations are: 't', 'f', 'true', or 'false'. If `strict` is `True`, only those representations
        are allowed, otherwise any non-False representation will return `True`.

        Args:
            x: the value
            strict: allow alternative non-False representations
            keys: keys of the currently validated element

        Returns:
            the boolean value represented by `x`

        Raises:
            ValidationError: if `x` does not represent a bool
        """
        if isinstance(x, bool):
            return x
        if isinstance(x, str):
            if x.lower() in ['f', 'false']:
                return False
            if not strict or x.lower() in ['t', 'true']:
                return True
        raise cls.ValidationError(f'Invalid text_bool value: {x} {cls.format_keys(keys)}')

    @classmethod
    def _execute_callable(cls, x, data_type, *args, strict, keys, **kwargs):
        """
        Execute a callable, and catch any exceptions. The result of the call to the callable is returned.

        Args:
            x:
            data_type:
            *args:
            strict:
            keys:
            **kwargs:

        Returns:
            The function results of the call to `x` with the provided arguments

        Raises:
            ValidationError if the call to `x` raises an exception, containing the exception message
        """
        try:
            return cls._fc(data_type, x, *args, strict=strict, keys=keys, **kwargs)
        except cls.ValidationError as e:
            raise e
        except Exception as e:
            if hasattr(data_type, '__name__'):
                name = data_type.__name__
            else:
                name = 'unnamed function {data_type}'
            raise cls.ValidationError(f'value at {cls.format_keys(keys)} passed to `{name}` raised `{e}`')

    @classmethod
    def _fc(cls, f, *args, **kwargs):
        """
        Helper function that calls a function, and removes `strict` and `keys` arguments if the function does not
        support them (and the class is allowed to use inspect).

        Args:
            f: the callable
            *args: arguments to pass to the callable
            **kwargs: keyword arguments to pass to the callable

        Returns:
            The return value of the callable, as called with the arguments
        """
        if cls.ALLOW_INSPECT:
            import inspect
            i_args = inspect.getargs(f.__code__).args
            if 'keys' in kwargs and 'keys' not in i_args:
                del kwargs['keys']
            if 'strict' in kwargs and 'strict' not in i_args:
                del kwargs['strict']
        return f(*args, **kwargs)

    @classmethod
    def validated(cls, x: Any, data_type: Any, strict: bool = True, _keys: Union[List, None] = None) -> Any:
        """
        Validate a value against a type specification. This function is called recursively, but typically it is called
        directly with `x` and `data_type` as dictionary arguments. Strict validation implies that only keys in the
        `data_type` are allowed in `x`, and that value types must match the defined type. If `strict` is `False`, then
        additional keys will be skipped, and types will be coerced to the defined type if possible.

        Args:
            x: the value to validate, typically a dict
            data_type: the type specification, typically a dict
            strict: whether to validate strictly
            _keys: for internal use in recursion

        Returns:
            The validated value, the same type as `x`, but possibly with some type coercion. The validated value will
            be newly constructed, using the same types as the original. Both dictionary and list types are supported
            for structuring the data.
        """
        if _keys is None:
            _keys = []
        if isinstance(data_type, dict):
            if not isinstance(x, dict):
                raise cls.ValidationError(f'expected {cls.format_keys(_keys)} to be a dict, not a {type(x)}')
            # if a constructor override was provided for type of x, call that with x, instead of the type constructor
            for t in cls.type_factories:
                if isinstance(x, t):
                    result = cls.type_factories[t](x)
                    # no check for multiple matches, first hit wins
                    break
            else:
                result = type(x)()
            wildcards = {
                key: dt for key, dt in data_type.items()
                if callable(key) or isinstance(key, tuple) or key.split(':')[0] == '*'
            }
            values = {}
            used_keys = []
            for key, type_value in data_type.items():
                if key not in wildcards:
                    new_keys = _keys + [key]
                    if not isinstance(key, str) and strict:
                        raise cls.ValidationError(f'non-string dictionary key {cls.format_keys(_keys)}')
                    # if the value is optional
                    if key not in x and (
                            (hasattr(type_value, '__origin__')
                             and type_value.__origin__ == Union
                             and type(None) in type_value.__args__) or
                            (isinstance(type_value, tuple)
                             and None in type_value)):
                        continue
                    if key not in x:
                        raise cls.ValidationError(f'missing required key {key} in {cls.format_keys(_keys)}')
                    if not wildcards:
                        result[key] = cls.validated(x[key], type_value, strict, new_keys)
                    else:
                        values[key] = (x[key], type_value, strict, new_keys)
                    used_keys.append(key)
            if wildcards:
                for key, value in x.items():
                    if key not in used_keys:
                        new_keys = _keys + [key]
                        for wildcard in wildcards:
                            try:
                                if isinstance(wildcard, type):
                                    if not isinstance(key, wildcard):
                                        if strict:
                                            raise cls.ValidationError(
                                                f'key {cls.format_keys(new_keys)} not of specified type {wildcard}')
                                        else:
                                            try:
                                                key = wildcard(key)
                                                new_keys[-1] = key
                                            except ValueError:
                                                raise cls.ValidationError(
                                                    f'key {cls.format_keys(new_keys)} cannot be cast to {wildcard}')
                                elif (
                                    (callable(wildcard) and cls._fc(wildcard, key, strict=strict, keys=new_keys) != key)
                                    or
                                    (isinstance(wildcard, tuple) and callable(wildcard[0]) and
                                     not isinstance(wildcard[0], type) and
                                     cls._fc(wildcard[0], key, *wildcard[1:], strict=strict, keys=new_keys) != key)
                                   ):
                                    raise cls.ValidationError(
                                        f'mismatch between key {cls.format_keys(new_keys)} and generated key')
                                elif (
                                    isinstance(wildcard, tuple) and
                                    (not callable(wildcard[0]) or isinstance(wildcard[0], type)) and
                                    key not in wildcard
                                   ):
                                    types = []
                                    converted = key
                                    for w in wildcard:
                                        if isinstance(w, type):
                                            types.append(w)
                                            try:
                                                if isinstance(key, w):
                                                    converted = key
                                                    break
                                                if id(key) == id(converted) and not strict:
                                                    converted = w(key)
                                            except ValueError:
                                                continue
                                    else:
                                        if id(key) == id(converted):
                                            if types:
                                                raise cls.ValidationError(
                                                    f'key {cls.format_keys(new_keys)} could not be cast to any {types}')
                                            else:
                                                raise cls.ValidationError(
                                                    f'restricted key {cls.format_keys(new_keys)} not in {wildcard}')
                                    key = converted
                                    new_keys[-1] = key
                                y = cls.validated(value, data_type[wildcard], strict, new_keys)
                                break
                            except cls.ValidationError as e:
                                if len(wildcards) == 1:
                                    raise cls.ValidationError(
                                        f'could not match to only wildcard {wildcard}, raised `{e}`')
                                continue
                        else:
                            if strict:
                                raise cls.ValidationError(
                                    f'value of {cls.format_keys(new_keys)} could not be matched to wildcard')
                            result[key] = x[key]
                            continue
                        result[key] = y
                    else:
                        result[key] = cls.validated(*values[key])
            return result
        elif hasattr(data_type, '__origin__') and (data_type.__origin__ == Union):
            result = cls.validated(x, tuple(data_type.__args__), strict, _keys)
        elif isinstance(data_type, tuple):
            if data_type and not isinstance(data_type[0], type) and callable(data_type[0]):
                result = cls._execute_callable(x, data_type[0], *data_type[1:], strict=strict, keys=_keys)
            else:
                for type_value in data_type:
                    # tuple may contain None when checking optional values
                    if type_value is None:
                        continue
                    try:
                        result = cls.validated(x, type_value, strict, _keys)
                        break
                    except cls.ValidationError as e:
                        if len(data_type) == 1:
                            raise e
                        continue
                else:
                    raise cls.ValidationError(
                        f'value does not match any of the optional types at {cls.format_keys(_keys)}')
        elif hasattr(data_type, '__origin__') and (data_type.__origin__ == list):
            result = cls.validated(x, list(data_type.__args__), strict, _keys)
        elif isinstance(data_type, list):
            # read an empty list as a list of any type (any list can be empty)
            data_type = [Any] if not data_type else data_type
            if not isinstance(x, list):
                raise cls.ValidationError(f'expected a list at {cls.format_keys(_keys)}, got {type(x)}')
            if len(data_type) != 1:
                raise cls.ValidationError(
                    f'multiple types for list content unexpected at {cls.format_keys(_keys)}: {data_type}')
            result = [cls.validated(elem, data_type[0], strict, _keys + [f'[{n}]']) for n, elem in enumerate(x)]
        elif isinstance(data_type, type):
            if isinstance(x, data_type):
                result = x
            else:
                if strict:
                    raise cls.ValidationError(f'strict type mismatch (no casting) at {cls.format_keys(_keys)}, '
                                              f'expected `{data_type.__name__}`, found `{type(x).__name__}`')
                try:
                    result = data_type(x)
                except ValueError:
                    raise cls.ValidationError(f'type mismatch, casting failed at {cls.format_keys(_keys)}, '
                                              f'expected `{data_type.__name__}`, found `{type(x).__name__}`')
        elif data_type is Any:
            result = x
        elif callable(data_type):
            result = cls._execute_callable(x, data_type, strict=strict, keys=_keys)
        elif x != data_type:
            # remaining option is identity, the "data_type" is a value that's required
            raise cls.ValidationError(f'value "{x}" does not match expected "{data_type}" at {cls.format_keys(_keys)}')
        else:
            result = copy(data_type)
        return result
